module Tweed
  ##
  # Abstract superclass to represent monads.
  #
  # To create a monad, subclass and define the monad’s Kleisli triple using
  # {Monad.define_monad define_monad}.
  # 
  # @example Implementing the Maybe monad
  #   class Maybe < Tweed::Monad
  #     define_monad do
  #       # This becomes the `initialize` instance method:
  #       construct { |value, success| @value, @success = value, success }
  #       
  #       # This becomes the `return` class method:
  #       unit { |value| self.new(value, true) }
  #       
  #       # This becomes the `bind` instance method:
  #       bind { |&block| @success ? block.call(@value) : self }
  #       
  #       # This becomes the (cached) `zero` class method:
  #       zero { self.new(nil, false) }
  #     end
  #     
  #     def inspect
  #       @success ? "Just #{@value.inspect}" : "Nothing"
  #     end
  #   end
  # 
  # @example Using the Maybe monad to implement safe division
  #   class SafeDivision < Maybe
  #     def /(other)
  #       self.bind do |x|
  #         other.bind do |y|
  #           y.zero? ? self.class.zero : self.class[x / y]
  #         end
  #       end
  #     end
  #   end
  #   
  #   x = SafeDivision[1] / SafeDivision[2.0] # => Just 0.5
  #   y = SafeDivision[1] / SafeDivision[0]   # => Nothing
  # 
  class Monad
    
    # Lift a value into this monad. 
    # @abstract Subclass and override using {Monad.define_monad define_monad}.
    # @return [Monad] the newly created monad.
    def self.return(*args)
      abstract
    end
    
    # Pull an underlying value out of this monad.
    # @abstract Subclass and override using {Monad.define_monad define_monad}.
    def bind(&block)
      abstract
    end
    
    # Syntactic sugar for {Monad.return Monad.return(...)}.
    # @return [Monad]
    def self.[](*args)
      self.return(*args)
    end
    
    ##
    # Adaptation of `liftM2` from Haskell. ‘Lifts’ a binary function into the
    # current monad.
    #
    # @example Add two numbers
    #   m_x = SomeMonad[1]
    #   m_y = SomeMonad[2]
    #   m_z = m_x.lift_m2(m_y) { |x, y| x + y } # => SomeMonad[1 + 2] => SomeMonad[3]
    #
    # @param other another value within the current monad.
    # @yield [x, y] the underlying values of `self` and `other`.
    # @return the result of the binary operation, lifted back into the current monad.
    def lift_m2(other)
      self.bind do |x|
        other.bind do |y|
          self.class[yield(x, y)]
        end
      end
    end
    
    protected
    
    ##
    # Class method to set up a Monad subclass using a Kleisli triple.
    #
    # Uses a custom DSL to define the triple and an optional monadic zero.
    #
    # @example Implementing the List monad
    #   class List < Tweed::Monad
    #     define_monad do
    #       # This becomes the `initialize` instance method, so instance
    #       # variables will be set on the Monad instance.
    #       construct { |sequence| @sequence = sequence }
    #       
    #       # This becomes the `return` class method, also accessible via `[]`
    #       # on both the class and instances.
    #       unit { |value| self.new([value]) }
    #       
    #       # This becomes the `bind` instance method, and will always receive
    #       # a block; it should explicitly use `block.call` instead of `yield`
    #       # due to the way blocks work in Ruby (Ruby will complain that no
    #       # block was passed if you `yield`).
    #       bind { |&block| self.new(@sequence.map(&block).inject(&:+)) }
    #       
    #       # The definition of zero for this monad. This will be called once
    #       # from the class and have its result cached.
    #       zero { self.new([].freeze!) }
    #     end
    #   end
    #
    # @see http://en.wikipedia.org/wiki/Monad_(functional_programming)#Concepts
    #   Wikipedia article on Monads
    def self.define_monad(&def_block)
      # Capture the Kleisli triple, and perhaps a monadic zero.
      capture = Object.new
      
      class << capture
        def construct(&block) @construct = block end
        def unit(&block)      @unit      = block end
        def bind(&block)      @bind      = block end
        def zero(&block)      @zero      = block end
      end
      
      capture.instance_eval(&def_block)
      
      class << capture
        def [](sym) instance_variable_get("@#{sym}".to_sym) end
      end
      
      # Define the necessary methods on the monad, given the components of the
      # Kleisli triple (and optionally the monadic zero).
      self.class_eval do
        define_method(:initialize, &capture[:construct]) unless capture[:construct].nil?
        define_method(:bind, &capture[:bind]) unless capture[:bind].nil?
        self.singleton_class.instance_eval do
          define_method(:return, &capture[:unit]) unless capture[:unit].nil?
          unless capture[:zero].nil?
            define_method(:get_zero, &capture[:zero])
            private(:get_zero)
            define_method(:zero) { @zero ||= get_zero }
          end
        end
      end
      
      self
    end # def self.define_monad
  end # class Monad
end # module Tweed

# @private
def abstract
  raise NotImplementedError
end